@@ -108,26 +108,17 @@ class Engine(object):
             raise Exception("can not set number of processes to %d" % n)
 
     @staticmethod
-    def run_async(f, enumerated_args, callback):
+    def run_async(f, args, callback, run_local=False):
         """ f is not implemented for handling enumerated items, so we exctract the items
         and construct a 'new enumeration' for the result.
         This is needed in order to ensure that the final result of the computations have
         a fixed order although their computation order might be different due to the
         concurrent processing."""
-        assert all(isinstance(arg, EnumeratedItem) for arg in enumerated_args)
-        for arg in enumerated_args:
-            if isinstance(arg.value, DelayedException):
-                callback(arg.value)
-                return
-        args = tuple(a for (__, a) in enumerated_args)
-        # dill can pickle more function types as pickle module (which underlies
-        # multiprocessing) , eg lambdas or static methods, so we transport our computation job
-        # serialized as strings, which we deserialize in the remote process again:
-        f_s = dill.dumps(f)
-        args_s = dill.dumps(args)
         # we use eval_pickled in both cases to make shure we have the same kind of error handling
         # for local and remote execution. this supports debugging.
-        if Engine.pool is None:
+        f_s = dill.dumps(f)
+        args_s = dill.dumps(args)
+        if run_local or Engine.pool is None:
             result = eval_pickled(f_s, args_s)
             callback(result)
         else:
@@ -141,25 +132,25 @@ class DataStream(object):
 
     def __init__(self):
         # self._stream = Queue.Queue()
-        self._streams = dict()
+        self._queues = dict()
 
-    def empty(self, consumer):
-        return self._streams.get(consumer).empty()
+    def empty(self, listener):
+        return self._queues.get(listener).empty()
 
-    def get(self, consumer):
-        return self._streams.get(consumer).get()
+    def get(self, listener):
+        return self._queues.get(listener).get()
 
     def size(self):
         return self._size
 
     def put(self, item):
         """this is where computation results are passed to."""
-        for stream in self._streams.values():
-            stream.put(item)
+        for queue in self._queues.values():
+            queue.put(item)
 
-    def register_consumer(self, consumer):
-        assert consumer not in self._streams
-        self._streams[consumer] = Queue.Queue()
+    def register_listener(self, listener):
+        assert listener not in self._queues
+        self._queues[listener] = Queue.Queue()
 
     def start_computations(self):
         """
@@ -187,8 +178,9 @@ class Source(DataStream):
 
 class Computation(DataStream):
 
-    def __init__(self, f, inputs):
+    def __init__(self, f, inputs, run_local=False):
         super(Computation, self).__init__()
+        self.run_local = run_local
         self.f = f
         self.input_streams = self._setup_inputs(inputs)
         self._size = self.compute_size()
@@ -202,7 +194,7 @@ class Computation(DataStream):
                     stream = Source(input_)
                 else:
                     stream = Source((input_,))
-            stream.register_consumer(self)
+            stream.register_listener(self)
             streams.append(stream)
         return streams
 
@@ -214,6 +206,16 @@ class Computation(DataStream):
     def _start_computation(self):
         pass
 
+    def _start_async_computation_for(self, result_number, args):
+        assert all(isinstance(arg, EnumeratedItem) for arg in args)
+        forward_result = wrap_callback(result_number, self.put)
+        for arg in args:
+            if isinstance(arg.value, DelayedException):
+                forward_result(arg.value)
+                return
+        args = tuple(a for (__, a) in args)
+        Engine.run_async(self.f, args, forward_result, self.run_local)
+
 
 class OutputNode(DataStream):
 
@@ -226,7 +228,7 @@ class OutputNode(DataStream):
             else:
                 input_ = Source((input_,))
 
-        input_.register_consumer(self)
+        input_.register_listener(self)
         self.input_ = input_
 
     def size(self):
@@ -238,7 +240,7 @@ class OutputNode(DataStream):
         """
         self.input_.start_computations()
 
-        results = [self.input_.get(self) for _ in range(self.size())]
+        results = [self.input_.get(listener=self) for _ in range(self.size())]
         if not do_not_reraise_exceptions:
             for result in results:
                 if isinstance(result.value, DelayedException):
@@ -292,13 +294,13 @@ class JoinComputation(Computation):
         while manager.input_data_pending(self.size()):
             started_computation = False
             for i, stream in enumerate(self.input_streams):
-                if not stream.empty(self):
-                    item = stream.get(self)
+                if not stream.empty(listener=self):
+                    item = stream.get(listener=self)
                     if isinstance(item.value, DelayedException):
                         self.put(item)
                         return
                     for number, args in manager.new_function_arguments_for(i, item):
-                        Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
+                        self._start_async_computation_for(number, args)
                         started_computation = True
             if not started_computation:
                 time.sleep(0.001)
@@ -317,7 +319,7 @@ class SummarizeComputation(Computation):
         items = []
         stream = self.input_streams[0]
         for i in range(stream.size()):
-            item = stream.get(self)
+            item = stream.get(listener=self)
             if isinstance(item.value, DelayedException):
                 self.put(item)
                 return
@@ -325,7 +327,7 @@ class SummarizeComputation(Computation):
         items.sort(key=lambda item: item.number)
         values = [v for __, v in items]
         arg = EnumeratedItem(0, values)
-        Engine.run_async(self.f, (arg,), callback=wrap_callback(0, self.put))
+        self._start_async_computation_for(0, (arg,))
 
 
 class ZipComputation(Computation):
@@ -350,12 +352,12 @@ class ZipComputation(Computation):
         # wait and get the constant args which we need anyway:
         cip = self._collect_data_from_constant_input_streams()
 
-        if not self._check_and_forward_exceptions(cip):
+        if self._found_and_handled_exception_in(cip):
             return
 
         # only constant inputs ?
         if self.size() == 1:
-            Engine.run_async(self.f, cip, callback=wrap_callback(0, self.put))
+            self._start_async_computation_for(0, cip)
         else:
             self._zip_all(cip)
 
@@ -363,16 +365,16 @@ class ZipComputation(Computation):
         constant_inputs = [None] * len(self.input_streams)
         for i, stream in enumerate(self.input_streams):
             if stream.size() == 1:
-                constant_inputs[i] = stream.get(self)
+                constant_inputs[i] = stream.get(listener=self)
         return constant_inputs
 
-    def _check_and_forward_exceptions(self, args):
+    def _found_and_handled_exception_in(self, args):
         for item in args:
             if item is not None:
                 if isinstance(item.value, DelayedException):
                     self.put(item)
-                    False
-        return True
+                    True
+        return False
 
     def _zip_all(self, cip):
         seen = set()
@@ -382,9 +384,9 @@ class ZipComputation(Computation):
             avail = self._check_for_any_new_data()
             if avail:
                 for number, args in self._assemble_function_arguments(input_matrix, seen):
-                    if not self._check_and_forward_exceptions(args):
+                    if self._found_and_handled_exception_in(args):
                         return
-                    Engine.run_async(self.f, args, callback=wrap_callback(number, self.put))
+                    self._start_async_computation_for(number, args)
                     n += 1
             else:
                 time.sleep(0.001)
@@ -400,14 +402,14 @@ class ZipComputation(Computation):
     def _check_for_any_new_data(self):
         for stream in self.input_streams:
             if stream.size() > 1:
-                if not stream.empty(self):
+                if not stream.empty(listener=self):
                     return True
         return False
 
     def _assemble_function_arguments(self, input_matrix, seen):
         for i, stream in enumerate(self.input_streams):
-            if not stream.empty(self):
-                item = stream.get(self)
+            if not stream.empty(listener=self):
+                item = stream.get(listener=self)
                 input_matrix[item.number][i] = item
 
         for ni, arg in enumerate(input_matrix):
@@ -425,6 +427,13 @@ def apply(inner):
     return wrapped
 
 
+def apply_local(inner):
+    @functools.wraps(inner)
+    def wrapped(*args):
+        return ZipComputation(inner, args, run_local=True)
+    wrapped.inner = inner
+    return wrapped
+
 def join(inner):
     @functools.wraps(inner)
     def wrapped(*args):
@@ -432,6 +441,12 @@ def join(inner):
     wrapped.inner = inner
     return wrapped
 
+def join_local(inner):
+    @functools.wraps(inner)
+    def wrapped(*args):
+        return JoinComputation(inner, args, run_local=True)
+    wrapped.inner = inner
+    return wrapped
 
 def summarize(inner):
     @functools.wraps(inner)
@@ -440,6 +455,12 @@ def summarize(inner):
     wrapped.inner = inner
     return wrapped
 
+def summarize_local(inner):
+    @functools.wraps(inner)
+    def wrapped(*args):
+        return SummarizeComputation(inner, args, run_local=True)
+    wrapped.inner = inner
+    return wrapped
 
 def output(inner):
     if isinstance(inner, Computation):
